@@ -96,6 +96,7 @@ enum state {
   STATE_FRESHENS,
   STATE_SOURCERPM,
   STATE_HEADERRANGE,
+  STATE_BUILDHOST,
 
   STATE_PROVIDESENTRY,
   STATE_REQUIRESENTRY,
@@ -196,6 +197,7 @@ static struct solv_xmlparser_element stateswitches[] = {
   { STATE_SOLVABLE,    "rpm:freshens",    STATE_FRESHENS,     0 },
   { STATE_SOLVABLE,    "rpm:sourcerpm",   STATE_SOURCERPM,    1 },
   { STATE_SOLVABLE,    "rpm:header-range", STATE_HEADERRANGE, 0 },
+  { STATE_SOLVABLE,    "rpm:buildhost",   STATE_BUILDHOST,    1 },
   { STATE_SOLVABLE,    "file",            STATE_FILE, 1 },
   { STATE_SOLVABLE,    "changelog",       STATE_CHANGELOG, 1 },
 
@@ -603,17 +605,20 @@ static void
 fill_cshash_from_new_solvables(struct parsedata *pd)
 {
   Pool *pool = pd->pool;
-  Id cstype = 0;
-  unsigned const char *cs;
-  int i;
+  int i, l;
+  KeyValue kv;
+  Repokey *key;
+
+  memset(&kv, 0, sizeof(kv));
 
   for (i = pd->first; i < pool->nsolvables; i++)
     {
       if (pool->solvables[i].repo != pd->repo)
 	continue;
-      cs = repodata_lookup_bin_checksum_uninternalized(pd->data, i, SOLVABLE_CHECKSUM, &cstype);
-      if (cs)
-	put_in_cshash(pd, cs, solv_chksum_len(cstype), i);
+      if ((key = repodata_lookup_kv_uninternalized(pd->data, i, SOLVABLE_CHECKSUM, &kv)) == 0)
+	continue;
+      if ((l = solv_chksum_len(key->type)) != 0)
+	put_in_cshash(pd, (const unsigned char *)kv.str, l, i);
     }
 }
 
@@ -1035,6 +1040,10 @@ endElement(struct solv_xmlparser *xmlp, int state, char *content)
       if (*content)
 	repodata_set_poolstr(pd->data, handle, SOLVABLE_PACKAGER, content);
       break;
+    case STATE_BUILDHOST:
+      if (*content)
+	repodata_set_str(pd->data, handle, SOLVABLE_BUILDHOST, content);
+      break;
     case STATE_SOURCERPM:
       if (*content)
 	repodata_set_sourcepkg(pd->data, handle, content);
@@ -1090,13 +1099,6 @@ endElement(struct solv_xmlparser *xmlp, int state, char *content)
     }
 }
 
-static void
-errorCallback(struct solv_xmlparser *xmlp, const char *errstr, unsigned int line, unsigned int column)
-{
-  struct parsedata *pd = xmlp->userdata;
-  pd->ret = pool_error(pd->pool, -1, "repo_rpmmd: %s at line %u:%u", errstr, line, column);
-}
-
 
 /*-----------------------------------------------*/
 
@@ -1134,8 +1136,9 @@ repo_add_rpmmd(Repo *repo, FILE *fp, const char *language, int flags)
       fill_cshash_from_repo(&pd);
     }
 
-  solv_xmlparser_init(&pd.xmlp, stateswitches, &pd, startElement, endElement, errorCallback);
-  solv_xmlparser_parse(&pd.xmlp, fp);
+  solv_xmlparser_init(&pd.xmlp, stateswitches, &pd, startElement, endElement);
+  if (solv_xmlparser_parse(&pd.xmlp, fp) != SOLV_XMLPARSER_OK)
+    pd.ret = pool_error(pool, -1, "repo_rpmmd: %s at line %u:%u", pd.xmlp.errstr, pd.xmlp.line, pd.xmlp.column);
   solv_xmlparser_free(&pd.xmlp);
 
   solv_free(pd.lastdirstr);
@@ -1144,6 +1147,22 @@ repo_add_rpmmd(Repo *repo, FILE *fp, const char *language, int flags)
   repodata_free_dircache(data);
   queue_free(&pd.diskusageq);
 
+  if ((flags & REPO_EXTEND_SOLVABLES) != 0)
+    {
+      /* is this a filelist extension? */
+      if (repodata_has_keyname(data, SOLVABLE_FILELIST))
+	repodata_set_filelisttype(data, REPODATA_FILELIST_EXTENSION);
+    }
+  else
+    {
+      /* is this a primary with a filtered filelist? */
+      if (data->end > data->start)
+	{
+	  repodata_set_filelisttype(data, REPODATA_FILELIST_FILTERED);
+	  repodata_set_void(data, SOLVID_META, REPOSITORY_FILTEREDFILELIST);
+	}
+    }
+
   if (!(flags & REPO_NO_INTERNALIZE))
     repodata_internalize(data);
   POOL_DEBUG(SOLV_DEBUG_STATS, "repo_add_rpmmd took %d ms\n", solv_timems(now));
