@@ -8,7 +8,7 @@
 /*
  * decision.c
  *
- * solver decision introspection code
+ * solver decision and alternative introspection code
  */
 
 #include <stdio.h>
@@ -322,6 +322,50 @@ move_decision(Queue *q, int to, int from)
   queue_deleten(q, from + 8, 8); 
 }
 
+/* sort a block of SOLVER_REASON_UNIT_RULE decisions */
+static void
+sort_unit_decisions(Solver *solv, Queue *q, int start, int end, Map *m)
+{
+  Pool *pool = solv->pool;
+  int i, j, k, doing = 1;
+  if (start + 8 == end)
+    {
+      Id truelit = q->elements[start];
+      MAPSET(m, truelit >= 0 ? truelit : -truelit);
+      return;
+    }
+  /* alternate between conflicts and installs, starting with conflicts */
+  for (i = start; i < end; doing ^= 1)
+    {
+      for (j = k = i; j < end; j += 8)
+	{
+	  Rule *or;
+	  Id p, pp;
+	  Id truelit = q->elements[j];
+	  if ((doing == 0 && truelit < 0) || (doing != 0 && truelit >= 0))
+	    continue;
+	  or = solv->rules + q->elements[j + 2];
+	  FOR_RULELITERALS(p, pp, or)
+	    if (p != truelit && !MAPTST(m, p >= 0 ? p : -p))
+	      break;
+	  if (p)
+	    continue;	/* not unit yet */
+	  if (j > k)
+	    move_decision(q, k, j);
+	  k += 8;
+	}
+      if (k == i)
+	continue;
+      if (i + 8 < k)
+	solv_sort(q->elements + i, (k - i) / 8, 8 * sizeof(Id), decisionsort_cmp, solv);
+      for (; i < k; i += 8)
+	{
+	  Id truelit = q->elements[i];
+	  MAPSET(m, truelit >= 0 ? truelit : -truelit);
+	}
+    }
+}
+
 static void
 solver_get_proof(Solver *solv, Id id, int flags, Queue *q)
 {
@@ -329,8 +373,7 @@ solver_get_proof(Solver *solv, Id id, int flags, Queue *q)
   Map seen, seent;	/* seent: was the literal true or false */
   Id rid, truelit;
   int first = 1;
-  int why, i, j, k;
-  int cnt, doing;
+  int why, i, cnt;
 
   queue_empty(q);
   if ((flags & SOLVER_DECISIONLIST_TYPEMASK) == SOLVER_DECISIONLIST_PROBLEM)
@@ -404,17 +447,19 @@ solver_get_proof(Solver *solv, Id id, int flags, Queue *q)
   queue_deleten(q, 0, cnt);
 
   /* switch last two decisions if the unsolvable rule is of type SOLVER_RULE_RPM_SAME_NAME */
-  if (q->count >= 16 && q->elements[q->count - 8 + 3] == SOLVER_RULE_RPM_SAME_NAME && q->elements[q->count - 16] > 0)
+  if (q->count >= 16 && q->elements[q->count - 8 + 4] == SOLVER_RULE_RPM_SAME_NAME && q->elements[q->count - 16] > 0)
     {
-      Rule *r = solv->rules + q->elements[q->count - 8 + 1];
+      Rule *r = solv->rules + q->elements[q->count - 8 + 2];
       /* make sure that the rule is a binary conflict and it matches the installed element */
       if (r->p < 0 && (r->d == 0 || r->d == -1) && r->w2 < 0
-	 && (q->elements[q->count - 16] == -r->p || q->elements[q->count - 16] -r->w2))
+	 && (q->elements[q->count - 16] == -r->p || q->elements[q->count - 16] == -r->w2))
 	{
-	  /* looks good! swap decisions and fixup truelit entries */
+	  /* looks good! swap decisions and fixup truelit/reason entries */
 	  move_decision(q, q->count - 16, q->count - 8);
 	  q->elements[q->count - 16] = -q->elements[q->count - 8];
+	  q->elements[q->count - 16 + 1] = SOLVER_REASON_UNIT_RULE;
 	  q->elements[q->count - 8] = 0;
+	  q->elements[q->count - 8 + 1] = SOLVER_REASON_UNSOLVABLE;
 	}
     }
 
@@ -444,42 +489,7 @@ solver_get_proof(Solver *solv, Id id, int flags, Queue *q)
       /* sort premise block */
       if (i > 8)
 	solv_sort(q->elements, i / 8, 8 * sizeof(Id), decisionsort_cmp, solv);
-
-      /* move rules:
-       *   if a package is installed, move conflicts right after
-       *   if a package is conflicted, move requires right after
-       */
-      doing = 1;
-      while (i < q->count - 8)
-	{
-	  doing ^= 1;
-	  for (j = k = i; j < q->count - 8; j += 8)
-	    {
-	      Rule *or;
-	      Id p, pp;
-	      Id truelit = q->elements[j];
-	      if ((doing == 0 && truelit < 0) || (doing != 0 && truelit >= 0))
-		continue;
-	      or = solv->rules + q->elements[j + 2];
-	      FOR_RULELITERALS(p, pp, or)
-		if (p != truelit && !MAPTST(&seen, p >= 0 ? p : -p))
-		  break;
-	      if (p)
-		continue;	/* not unit yet */
-	      if (j > k)
-		move_decision(q, k, j);
-	      k += 8;
-	    }
-	  if (k == i)
-	    continue;
-	  if (i + 8 < k)
-	    solv_sort(q->elements + i, (k - i) / 8, 8 * sizeof(Id), decisionsort_cmp, solv);
-	  for (; i < k; i += 8)
-	    {
-	      Id truelit = q->elements[i];
-	      MAPSET(&seen, truelit >= 0 ? truelit : -truelit);
-	    }
-	}
+      sort_unit_decisions(solv, q, i, q->count - 8, &seen);
     }
 
   map_free(&seen);
@@ -667,7 +677,7 @@ getdecisionlist(Solver *solv, Map *dm, int flags, Queue *decisionlistq)
 	      Id p2 = iq.elements[i];
 	      if (p2 < 0)
 		MAPSET(dm, -p2);
-	      else if (solv->decisionmap[p2] > 0)
+	      else if (reason == SOLVER_REASON_UNIT_RULE || solv->decisionmap[p2] > 0)
 		MAPSET(dm, p2);
 	    }
 	  break;
@@ -676,8 +686,40 @@ getdecisionlist(Solver *solv, Map *dm, int flags, Queue *decisionlistq)
 	}
     }
   queue_free(&iq);
-  if ((flags & (SOLVER_DECISIONLIST_SORTED | SOLVER_DECISIONLIST_WITHINFO)) == SOLVER_DECISIONLIST_SORTED)
+  if ((flags & SOLVER_DECISIONLIST_SORTED) != 0)
     {
+      /* reuse dm as "seen" map */
+      MAPZERO(dm);
+      MAPSET(dm, 1);
+      for (i = 0; i < decisionlistq->count; i += 8)
+	{
+	  if (decisionlistq->elements[i + 1] != SOLVER_REASON_UNIT_RULE)
+	    {
+	      Id p = decisionlistq->elements[i] < 0 ? -decisionlistq->elements[i] : decisionlistq->elements[i];
+	      MAPSET(dm, p);
+	    }
+	  else
+	    {
+	      int j;
+	      for (j = i + 8; j < decisionlistq->count; j += 8)
+		if (decisionlistq->elements[j + 1] != SOLVER_REASON_UNIT_RULE)
+		  break;
+	      sort_unit_decisions(solv, decisionlistq, i, j, dm);
+	      i = j - 8;
+	    }
+	}
+    }
+  if ((flags & SOLVER_DECISIONLIST_WITHINFO) != 0)
+    {
+      /* set decisioninfo bits */
+      for (i = 0; i < decisionlistq->count; i += 8)
+	decisionlistq->elements[i + 3] = solver_calc_decisioninfo_bits(solv, decisionlistq->elements[i], decisionlistq->elements[i + 4], decisionlistq->elements[i + 5], decisionlistq->elements[i + 6], decisionlistq->elements[i + 7]);
+      if (flags & SOLVER_DECISIONLIST_MERGEDINFO)
+	decisionmerge(solv, decisionlistq);
+    }
+  else if ((flags & SOLVER_DECISIONLIST_SORTED) != 0)
+    {
+      /* strip the info elements we added for sorting */
       int j;
       for (i = j = 0; i < decisionlistq->count; i += 8)
 	{
@@ -687,14 +729,6 @@ getdecisionlist(Solver *solv, Map *dm, int flags, Queue *decisionlistq)
 	}
       queue_truncate(decisionlistq, j);
     }
-  else
-    {
-      /* set decisioninfo bits */
-      for (i = 0; i < decisionlistq->count; i += 8)
-	decisionlistq->elements[i + 3] = solver_calc_decisioninfo_bits(solv, decisionlistq->elements[i], decisionlistq->elements[i + 4], decisionlistq->elements[i + 5], decisionlistq->elements[i + 6], decisionlistq->elements[i + 7]);
-      if (flags & SOLVER_DECISIONLIST_MERGEDINFO)
-	decisionmerge(solv, decisionlistq);
-    }
 }
 
 void
@@ -707,6 +741,7 @@ solver_get_decisionlist(Solver *solv, Id id, int flags, Queue *decisionlistq)
   map_init(&dm, pool->nsolvables);
   MAPSET(&dm, id);
   getdecisionlist(solv, &dm, flags, decisionlistq);
+  map_free(&dm);
   if (!decisionlistq->count)
     {
       queue_push(decisionlistq, -id);
@@ -718,7 +753,6 @@ solver_get_decisionlist(Solver *solv, Id id, int flags, Queue *decisionlistq)
 	  queue_push2(decisionlistq, 0, 0);
 	}
     }
-  map_free(&dm);
 }
 
 void
@@ -738,6 +772,7 @@ solver_get_decisionlist_multiple(Solver *solv, Queue *idq, int flags, Queue *dec
 	MAPSET(&dm, p);
     }
   getdecisionlist(solv, &dm, flags, decisionlistq);
+  map_free(&dm);
   for (i = 0; i < idq->count; i++)
     {
       Id p = idq->elements[i];
@@ -752,7 +787,6 @@ solver_get_decisionlist_multiple(Solver *solv, Queue *idq, int flags, Queue *dec
 	  queue_push2(decisionlistq, 0, 0);
 	}
     }
-  map_free(&dm);
 }
 
 
@@ -1008,3 +1042,190 @@ solver_decisioninfo2str(Solver *solv, int bits, int type, Id from, Id to, Id dep
       return "unsupported decision merge?";
   return solver_ruleinfo2str(solv, type, from, to, dep);
 }
+
+int
+solver_alternatives_count(Solver *solv)
+{
+  Id *elements = solv->branches.elements;
+  int res, count;
+  for (res = 0, count = solv->branches.count; count; res++)
+    count -= elements[count - 2];
+  return res;
+}
+
+int
+solver_get_alternative(Solver *solv, Id alternative, Id *idp, Id *fromp, Id *chosenp, Queue *choices, int *levelp)
+{
+  int cnt = solver_alternatives_count(solv);
+  int count = solv->branches.count;
+  Id *elements = solv->branches.elements;
+  if (choices)
+    queue_empty(choices);
+  if (alternative <= 0 || alternative > cnt)
+    return 0;
+  elements += count;
+  for (; cnt > alternative; cnt--)
+    elements -= elements[-2];
+  if (levelp)
+    *levelp = elements[-1];
+  if (fromp)
+    *fromp = elements[-4];
+  if (idp)
+    *idp = elements[-3];
+  if (chosenp)
+    {
+      int i;
+      *chosenp = 0;
+      for (i = elements[-2]; i > 4; i--)
+	{
+	  Id p = -elements[-i];
+	  if (p > 0 && solv->decisionmap[p] == elements[-1] + 1)
+	    {
+	      *chosenp = p;
+	      break;
+	    }
+	}
+    }
+  if (choices)
+    queue_insertn(choices, 0, elements[-2] - 4, elements - elements[-2]);
+  return elements[-4] ? SOLVER_ALTERNATIVE_TYPE_RECOMMENDS : SOLVER_ALTERNATIVE_TYPE_RULE;
+}
+
+static int
+find_alternative_rule_from_learnt_rec(Solver *solv, int rid, Map *m, int cnt)
+{
+  Pool *pool = solv->pool;
+  int why = solv->learnt_why.elements[rid - solv->learntrules];
+  while ((rid = solv->learnt_pool.elements[why++]) != 0)
+    {
+      Rule *r = solv->rules + rid;
+      Id p, pp;
+      int c;
+      if (rid >= solv->learntrules)
+	{
+	  if ((rid = find_alternative_rule_from_learnt_rec(solv, rid, m, cnt)))
+	    return rid;
+	  continue;
+	}
+      c = 0;
+      FOR_RULELITERALS(p, pp, r)
+	if (p > 0 && MAPTST(m, p))
+	  c++;
+      if (c == cnt)	/* all bits hit */
+	return rid;
+    }
+  return 0;
+}
+
+static int
+find_alternative_rule_from_learnt(Solver *solv, int rid)
+{
+  Pool *pool = solv->pool;
+  Map m;
+  int i, count, cnt;
+  Id *elements = solv->branches.elements;
+
+  /* find alternative by rule id */
+  for (count = solv->branches.count; count; count -= elements[count - 2])
+    if (elements[count - 4] == 0 && elements[count - 3] == rid)
+      break;
+  if (!count)
+    return 0;
+  map_init(&m, pool->nsolvables);
+  cnt = 0;
+  for (i = count - elements[count - 2]; i < count - 4; i++)
+    if (elements[i] > 0)
+      {
+        MAPSET(&m, elements[i]);
+        cnt++;
+      }
+  rid = find_alternative_rule_from_learnt_rec(solv, rid, &m, cnt);
+  map_free(&m);
+  return rid;
+}
+
+int
+solver_alternativeinfo(Solver *solv, int type, Id id, Id from, Id *fromp, Id *top, Id *depp)
+{
+  if (fromp)
+    *fromp = 0;
+  if (top)
+    *top = 0;
+  if (depp)
+    *depp = 0;
+  if (type == SOLVER_ALTERNATIVE_TYPE_RECOMMENDS)
+    {
+      if (fromp)
+        *fromp = from;
+      if (depp)
+        *depp = id;
+      return SOLVER_RULE_PKG_RECOMMENDS;
+    }
+  else if (type == SOLVER_ALTERNATIVE_TYPE_RULE)
+    {
+      int rclass = solver_ruleclass(solv, id);
+      if (rclass == SOLVER_RULE_LEARNT)
+	{
+	  id = find_alternative_rule_from_learnt(solv, id);
+	  if (!id)
+	    return SOLVER_RULE_LEARNT;
+          rclass = solver_ruleclass(solv, id);
+	}
+      if (rclass == SOLVER_RULE_CHOICE || rclass == SOLVER_RULE_RECOMMENDS)
+	id = solver_rule2pkgrule(solv, id);
+      else if (rclass == SOLVER_RULE_BEST)
+	{
+	  Id info = solv->bestrules_info[id - solv->bestrules];
+	  if (info > 0)
+	    {
+	      /* best update */
+	      if (fromp)
+		*fromp = info;
+	      return SOLVER_RULE_UPDATE;
+	    }
+	  id = -info;		/* best job, delegate to job rule */
+	}
+      return solver_ruleinfo(solv, id, fromp, top, depp);
+    }
+  return 0;
+}
+
+const char *
+solver_alternative2str(Solver *solv, int type, Id id, Id from)
+{
+  const char *s;
+  Pool *pool = solv->pool;
+  Id to, dep;
+  type = solver_alternativeinfo(solv, type, id, from, &from, &to, &dep);
+  switch (type)
+    {
+    case SOLVER_RULE_PKG_RECOMMENDS:
+      s = pool_dep2str(pool, dep);
+      if (from)
+        s =  pool_tmpappend(pool, s, ", recommended by ", pool_solvid2str(pool, from));
+      return s;
+    case SOLVER_RULE_PKG_REQUIRES:
+      s = pool_dep2str(pool, dep);
+      if (from)
+        s =  pool_tmpappend(pool, s, ", required by ", pool_solvid2str(pool, from));
+      return s;
+    case SOLVER_RULE_PKG_CONFLICTS:
+      s = pool_dep2str(pool, dep);
+      if (from)
+        s =  pool_tmpappend(pool, s, ", conflicted by  ", pool_solvid2str(pool, from));
+      return s;
+    case SOLVER_RULE_YUMOBS:
+      return pool_tmpjoin(pool, pool_id2str(pool, pool->solvables[to].name), ", obsoleting ", pool_dep2str(pool, dep));;
+    case SOLVER_RULE_JOB:
+      if ((to & SOLVER_SELECTMASK) == SOLVER_SOLVABLE_PROVIDES || (to & SOLVER_SELECTMASK) == SOLVER_SOLVABLE_NAME)
+	return pool_dep2str(pool, dep);
+      return solver_select2str(pool, to & SOLVER_SELECTMASK, dep);
+    case SOLVER_RULE_UPDATE:
+    case SOLVER_RULE_FEATURE:
+      return pool_solvid2str(pool, from);
+    default:
+      break;
+    }
+  return solver_ruleinfo2str(solv, type, from, to, dep);
+}
+
